// Original plugin by Franck Gartemann
// Extended version by MVV

library Expander2;

uses
    SysUtils, Windows, Classes, Inifiles;

{$E wdx}

const
    ft_nomorefields = 0;
    ft_numeric_32 = 1;
    ft_numeric_64 = 2;
    ft_numeric_floating = 3;
    ft_date = 4;
    ft_time = 5;
    ft_boolean = 6;
    ft_multiplechoice = 7;
    ft_string = 8;
    ft_fulltext = 9;
    ft_datetime = 10;

    //For ContentGetValue
    ft_nosuchfield = -1;
    ft_fileerror = -2;
    ft_fieldempty = -3;
    ft_ondemand = -4;
    ft_delayed = 0;
    CONTENT_DELAYIFSLOW = 1;  // ContentGetValue called in foreground

var
    FieldCount: Integer; // Number of fields
    seps: TStringList;


{------------------------------------------------------------------------------}


function ContentGetSupportedField(FieldIndex: Integer; FieldName: PChar; Units: PChar; MaxLen: Integer): Integer; stdcall;
begin
    result := 0;
    if ((FieldIndex < 0) or (FieldIndex > fieldcount - 1)) then Result := ft_nomorefields;
    if ((FieldIndex >= 0) and (FieldIndex < fieldcount)) then
    begin
        StrCopy(Units, PChar('FromName|FromFullName'));
        StrCopy(FieldName, PChar('Part' + IntToStr(FieldIndex+1))); //"Telling" to TC what field we will show :)
        result := ft_string;
    end;
end;


{------------------------------------------------------------------------------}


procedure SplitString(Input: String; var Output: TStringList);
var
   i, j: Integer;

begin
     Output.Clear;
     i := 1;
     while (i <= Length(Input)) do begin
           j := i;
           while (j <= Length(Input)) and (Input[j] <> '|') do inc(j);
           if (j > i) then Output.Add(Copy(Input, i, j-i));
           i := j + 1;
     end;
end;

function AtPosition(Str, SubStr: String; Pos: Integer): Boolean;
var
   i: Integer;
begin
     if (Length(Str) - Pos < Length(SubStr)) then result := false
     else begin
          i := 1;
          while (i <= Length(SubStr)) and (Str[Pos + i] = SubStr[i]) do inc(i);
          if (i > Length(SubStr)) then result := true else result := false;
     end;
end;

procedure GetSubstring(Input: String; Delims: TStringList; Index: Integer; var Output: String);
var
   i, j, k, n: Integer;
begin
     Output := '';
     i := 1;
     while (i <= Length(Input)) do begin
          n := 0;
          j := i;
          while (j <= Length(Input)) do begin
              for k := 0 to Delims.Count - 1 do if (AtPosition(Input, Delims[k], j)) then begin
                  n := Length(Delims[k]);
                  break;
              end;
              if (n > 0) then break;
              inc(j);
          end;

          if (n > 0) or (j >= Length(Input)) then dec(Index);

          if (Index < 0) then begin
             Output := Copy(Input, i, j-i+1);
             break;
          end;

          i := j + n + 1;
     end;
end;


function ContentGetValue(FilePath: PChar; FieldIndex, UnitIndex: Integer; FieldValue: PChar; MaxLen, Flags: Integer): Integer; stdcall;
var
   s, t: String;
begin
    if ((FieldIndex > -1) and (FieldIndex < fieldcount)) then begin
        if (UnitIndex = 0) then s := ExtractFileName(FilePath) else s := FilePath;

        GetSubstring(s, seps, 0, t);
        GetSubstring(s, seps, 1, t);
        GetSubstring(s, seps, 2, t);

        GetSubstring(s, seps, FieldIndex, t);
        if (Length(t) = 0) then result := ft_fieldempty
        else begin
           StrCopy(PChar(FieldValue), PChar(t));
           result := ft_string;
        end;
    end
    else result := ft_fileerror;
end;


{------------------------------------------------------------------------------}


exports
    ContentGetValue,
    ContentGetSupportedField;


{------------------------------------------------------------------------------}


procedure Init;
var
    path, sep: String;
    tmpIni: TIniFile;
    ModName: Array [0..259] of Char;

begin
    SetString(path, ModName, Windows.GetModuleFileName(HInstance, ModName, SizeOf(ModName)));
    path := ExtractFilePath(path) + 'Expander2.lng';
    tmpIni := TIniFile.Create(path);
    FieldCount := tmpIni.ReadInteger('Main', 'FieldCount', 64);
    sep := tmpIni.ReadString('Main', 'Divider', ' ');
    seps := TStringList.Create;
    seps.Capacity := 16;
    SplitString(sep, seps);
end;


{------------------------------------------------------------------------------}


begin
  Init;
end.