// Original plugin by Franck Gartemann
// Extended version by MVV
// Unicode and x64 version by ProgMan13

library Expander2;

{$R expander2.res}

{$IFDEF WIN64}
  {$E w64}
  {$IFNDEF VER230}
    {$E wdx64}
  {$ENDIF}
{$ENDIF}
{$IFDEF WIN32}
  {$E wdx}
{$ENDIF}

{$WEAKLINKRTTI ON}

uses
  SysUtils,
  Windows,
  Classes,
  Inifiles;

const
  DefSeparator = '" "';
  DefFieldCount = 64;

  ft_nomorefields = 0;
  ft_numeric_32 = 1;
  ft_numeric_64 = 2;
  ft_numeric_floating = 3;
  ft_date = 4;
  ft_time = 5;
  ft_boolean = 6;
  ft_multiplechoice = 7;
  ft_string = 8;
  ft_fulltext = 9;
  ft_datetime = 10;
  ft_StringW = 11;

  // For ContentGetValue
  ft_nosuchfield = -1;
  ft_fileerror = -2;
  ft_fieldempty = -3;
  ft_ondemand = -4;
  ft_NotSupported = -5;
  ft_delayed = 0;
  CONTENT_DELAYIFSLOW = 1; // ContentGetValue called in foreground

type
  TWordDelimiters = array of WideChar;

var
  FieldCount: Integer; // Number of fields
  seps: TStringList;
  WordDelimiters: TWordDelimiters;
  OldStyle, CaseSensitive: Boolean;

  { ------------------------------------------------------------------------------ }

function DequoteString(const Str: string; const QuoteChar: Char = '"')
  : string; inline; 
begin
  if Length(Str) < 2 then
    Exit(Str);
  if (Str[1] = QuoteChar) and (Str[Length(Str)] = QuoteChar) then
    Result := Copy(Str, 2, Length(Str) - 2)
  else
    Result := Str;
end;

function ContentGetSupportedField(FieldIndex: Integer; FieldName: PAnsiChar;
  Units: PAnsiChar; MaxLen: Integer): Integer; stdcall;
begin
  Result := 0;
  if ((FieldIndex < 0) or (FieldIndex > FieldCount - 1)) then
    Result := ft_nomorefields;
  if ((FieldIndex >= 0) and (FieldIndex < FieldCount)) then
  begin
    StrLCopy(Units, PAnsiChar('FromName|FromFullName|' +
      'FromNameWithoutExtension|FromFullNameWithoutExtension|FromPathOnly'),
      MaxLen - 1);

    if not OldStyle then
    begin
      if FieldCount - FieldIndex <= FieldCount div 2 then
        StrLCopy(FieldName,
          PAnsiChar(AnsiString('Part' + IntToStr(0 - (FieldCount - (FieldIndex))
          ))), MaxLen - 1)
      else
        StrLCopy(FieldName,
          PAnsiChar(AnsiString('Part' + IntToStr(FieldIndex + 1))), MaxLen - 1);
    end
    else
      StrLCopy(FieldName, PAnsiChar(AnsiString('Part' + IntToStr(FieldIndex + 1)
        )), MaxLen - 1);

    // "Telling" to TC what field we will show :)
    Result := ft_string;
  end;
end;

{ ------------------------------------------------------------------------------ }

procedure SplitString(const Input: string; const Output: TStringList);
var
  i, j: Integer;
begin
  Output.Clear;
  i := 1;
  while (i <= Length(Input)) do
  begin
    j := i;
    while (j <= Length(Input)) and (Input[j] <> '|') do
      inc(j);
    if (j > i) then
      Output.Add(Copy(Input, i, j - i));
    i := j + 1;
  end;
end;

function AtPosition(const Str, SubStr: string; const Pos: Integer)
  : Boolean;{$IFDEF WIN32} inline; {$ENDIF}
var
  i: Integer;
begin
  if (Length(Str) - Pos < Length(SubStr)) then
    Result := False
  else
  begin
    i := 1;
    while (i <= Length(SubStr)) and (Str[Pos + i] = SubStr[i]) do
      inc(i);
    if (i > Length(SubStr)) then
      Result := True
    else
      Result := False;
  end;
end;

procedure GetSubstring(const Input, Original: string; const Delims: TStringList;
  Index: Integer; var Output: string);{$IFDEF WIN32} inline; {$ENDIF}
var
  i, j, k, n: Integer;
begin
  Output := '';
  i := 1;
  while (i <= Length(Input)) do
  begin
    n := 0;
    j := i;
    while (j <= Length(Input)) do
    begin
      for k := 0 to Delims.Count - 1 do
        if (AtPosition(Input, Delims[k], j)) then
        begin
          n := Length(Delims[k]);
          Break;
        end;
      if (n > 0) then
        Break;
      inc(j);
    end;

    if (n > 0) or (j >= Length(Input)) then
      dec(index);

    if (index < 0) then
    begin
      Output := Copy(Original, i, j - i + 1);
      Break;
    end;

    i := j + n + 1;
  end;
end;

function CheckChar(const AChar: Char; const WordDelims: TWordDelimiters)
  : Boolean;{$IFDEF WIN32} inline; {$ENDIF}
var
  C: Char;
begin
  Result := False;
  for C in WordDelims do
    if C = AChar then
    begin
      Result := True;
      Break;
    end;
end;

function WordPosition(const N: Integer; const S: string;
  const WordDelims: TWordDelimiters): Integer;{$IFDEF WIN32} inline; {$ENDIF}
var
  Count, I: Integer;
begin
  Count := 0;
  I := 1;
  Result := 0;
  while (I <= Length(S)) and (Count <> N) do begin
    { skip over delimiters }
    while (I <= Length(S)) and CheckChar(S[I], WordDelims) do Inc(I);
    { if we're not beyond end of S, we're at the start of a word }
    if I <= Length(S) then Inc(Count);
    { if not finished, find the end of the current word }
    if Count <> N then
      while (I <= Length(S)) and not CheckChar(S[I], WordDelims) do Inc(I)
    else Result := I;
  end;
end;

function ExtractWord(N: Integer; const S, Original: string;
  const WordDelims: TWordDelimiters): string;{$IFDEF WIN32} inline; {$ENDIF}
var
  I: Integer;
  Len: Integer;
begin
  Len := 0;
  I := WordPosition(N, S, WordDelims);
  if I <> 0 then
    { find the end of the current word }
    while (I <= Length(S)) and not CheckChar(S[I], WordDelims) do begin
      { add the I'th character to result }
      Inc(Len);
      SetLength(Result, Len);
      Result[Len] := Original[I];
      Inc(I);
    end;
  SetLength(Result, Len);
end;

function WordCount(const S: string; const WordDelims: TWordDelimiters): Integer;{$IFDEF WIN32} inline; {$ENDIF}
var
  SLen, I: Cardinal;
begin
  Result := 0;
  I := 1;
  SLen := Length(S);
  while I <= SLen do begin
    while (I <= SLen) and CheckChar(S[I], WordDelims) do Inc(I);
    if I <= SLen then Inc(Result);
    while (I <= SLen) and not CheckChar(S[I], WordDelims) do Inc(I);
  end;
end;

function ContentGetValueW(FilePath: PWideChar; FieldIndex, UnitIndex: Integer;
  FieldValue: PByte; MaxLen, Flags: Integer): Integer; stdcall;
var
  S, t, Original: string;
  AWordsCount, CurrentField: Integer;
begin
  if ((FieldIndex > -1) and (FieldIndex < FieldCount)) then
  begin
    if (UnitIndex = 0) or (UnitIndex = 2) then
      S := ExtractFileName(FilePath)
    else if UnitIndex = 4 then
      S := ExtractFilePath(FilePath)
    else
      S := FilePath;

    if (UnitIndex = 2) or (UnitIndex = 3) then
    begin
      S := ChangeFileExt(S, '.');
      if S[Length(S)] = '.' then
        S := Copy(S, 1, Length(S) - 1);
    end;

    Original := S;
    if not CaseSensitive then
      S := AnsiLowerCase(S);

    if not OldStyle then
      if FieldCount - FieldIndex <= FieldCount div 2 then
      begin
        AWordsCount := WordCount(S, WordDelimiters);
        CurrentField := 0 - (FieldCount - (FieldIndex));
        if Abs(CurrentField) > AWordsCount then
          t := ''
        else
          t := ExtractWord(AWordsCount + CurrentField + 1, S, Original, WordDelimiters);
      end
      else
      begin
        t := ExtractWord(FieldIndex + 1, S, Original, WordDelimiters);
      end
    else
    begin
      //GetSubstring(S, Original, seps, 0, t);
      //GetSubstring(S, Original, seps, 1, t);
      //GetSubstring(S, Original, seps, 2, t);

      GetSubstring(S, Original, seps, FieldIndex, t);
    end;


    if (Length(t) = 0) then
      Result := ft_fieldempty
    else
    begin
      StrLCopy(PChar(FieldValue), PChar(t), MaxLen div 2 - 2);
      Result := ft_StringW;
    end;
  end
  else
    Result := ft_fileerror;
end;

function ContentGetValue(FilePath: PAnsiChar; FieldIndex, UnitIndex: Integer;
  FieldValue: PAnsiChar; MaxLen, Flags: Integer): Integer; stdcall;
begin
  Result := ft_NotSupported;
end;

{ ------------------------------------------------------------------------------ }

exports
  ContentGetValue,
  ContentGetValueW,
  ContentGetSupportedField;

{ ------------------------------------------------------------------------------ }

function CheckINI: string;
var
  INIFile, ExampleFile, sep: string;
  OldINI, NewINI: TMemIniFile;
  FCount: Integer;
begin
  INIFile := ExtractFilePath(GetModuleName(HInstance)) + 'Expander2.ini';
  ExampleFile := ExtractFilePath(GetModuleName(HInstance)) + 'Expander2.ini.example';

  if FileExists(INIFile) then
  begin
    Result := INIFile;
    Exit;
  end;

  Result := ExampleFile;

  try
    OldINI := TMemIniFile.Create(ExampleFile);
    try
      FCount := OldINI.ReadInteger('Main', 'FieldCount', DefFieldCount);
      sep := OldINI.ReadString('Main', 'Divider', DefSeparator);

      NewINI := TMemIniFile.Create(INIFile);
      try
        NewINI.WriteInteger('Main', 'FieldCount', FCount);
        NewINI.WriteString('Main', 'Divider', sep);
        NewINI.WriteBool('Main', 'OldStyle', False);
        NewINI.WriteBool('Main', 'CaseSensitive', True);
        NewINI.Encoding := TEncoding.Unicode;
        NewINI.UpdateFile;
      finally
        NewINI.Free;
      end;
    finally
      OldINI.Free;
    end;

    Result := INIFile;
  except
  end;
end;

procedure Init;
var
  path, sep: string;
  tmpIni: TMemIniFile;
  I: Integer;
begin
  path := CheckINI;
  try
    tmpIni := TMemIniFile.Create(path);
    try
      tmpIni.AutoSave := False;
      FieldCount := tmpIni.ReadInteger('Main', 'FieldCount', DefFieldCount);
      OldStyle := tmpIni.ReadBool('Main', 'OldStyle', False);
      CaseSensitive := tmpIni.ReadBool('Main', 'CaseSensitive', True);
      sep := tmpIni.ReadString('Main', 'Divider', DefSeparator);

      sep := DequoteString(sep);

      if not CaseSensitive then
        sep := AnsiLowerCase(sep);

    finally
      tmpIni.Free;
    end;
  except
    sep := DefSeparator;
  end;
  seps := TStringList.Create;
  seps.Capacity := 16;
  SplitString(sep, seps);

  SetLength(WordDelimiters, seps.Count);
  for I := 0 to seps.Count - 1 do
    WordDelimiters[I] := seps[I][1];
end;

{ ------------------------------------------------------------------------------ }

procedure DllInit(Code: Integer);
begin
  case Code of
    DLL_PROCESS_ATTACH:
      begin
        Init;
      end;
    DLL_PROCESS_DETACH:
      begin
        seps.Free;
      end;
  end; // case
end;

begin
  DisableThreadLibraryCalls(HInstance);
  DllProc := @DllInit;
  DllInit(DLL_PROCESS_ATTACH);
end.
